"""
SPK_SCAN3D        - first written in 2020

To the extent possible under law, the author(s) have dedicated all copyright and
related and neighboring rights to this software to the public domain worldwide.
This software is distributed without any warranty.

Revisions:
17-Mar-2020
Initial version
"""
 
import numpy as np
from scipy.special import spherical_jn, spherical_yn, sph_harm
from mayavi import mlab
import time

R_air = 287.058    # Specific gas constant of dry air 

def spherical_hn1(n, z):
    """ Spherical Hankel functions of the first kind. """
    return spherical_jn(n, z) + 1j * spherical_yn(n, z)

def spherical_hn2(n, z):
    """ Spherical Hankel functions of the second kind. """
    return spherical_jn(n, z) - 1j * spherical_yn(n, z)

def sph_to_cart(r, theta, phi):
    """ Spherical to Cartesian coordinates conversion.
        Inputs: r, theta (colatitude), phi (azimuth/longitude)
        Outputs: arrays of x, y, z.
    """
    return (np.asarray(r * np.sin(theta) * np.cos(phi)),
            np.asarray(r * np.sin(theta) * np.sin(phi)),
            np.asarray(r * np.cos(theta)))

def dB_p_abs(p):
    """ 20 * log10(magnitude(p)) - For log scale pressure magnitude plotting.
    """
    return 20 * np.log10(np.abs(p))

def conv_norm(p_n, p_n_minus_1):
    """ Calculate the reconstruction pressure convergence/error metric.
        Function is typically used for comparing the reconstructed pressures
        at the current spherical harmonics order those of a previous step.
        The normalized maximum difference (in magnitude) is returned.
        Inputs: p_n - pressures at order n (or reference pressures)
                p_n_minus_1 - pressures at order n-1 (or pressures to be
                              compared against the reference pressures)
                p_n and p_n_minus_1 are arrays of the same size
        Output: Max[Magnitude(P_n - P_n-1)] / Mean[Magnitude(P_n)]
    """
    return (np.abs(np.ravel(p_n - p_n_minus_1)).max() 
                                / np.abs(np.ravel(p_n)).mean())

def cart_to_sph(x, y, z):
    """ Cartesian to sppherical coordinates conversion.
        Inputs: x, y, z
        Outputs: arrays of r, theta (colatitude), phi (azimuth/longitude)
    """
    r = np.asarray(np.linalg.norm(np.array([x, y, z]), axis=0))
    theta = np.asarray(np.arccos(z / r))
    phi = np.asarray(np.arctan2(y, x))
    phi[phi < 0] = phi[phi < 0] + 2 * np.pi
    return r, theta, phi

def equi_ang_angles(delta_theta=np.pi/36, delta_phi=np.pi/36,
                    include_poles=False):
    """ Create the angular coordinates of a grid with equi-angular spacings
        in 'theta' (polar/colatitudinal) and 'phi' (azimuthal/longitudinal)
        directions.
        If 'include_poles' is False, the polar north and south poles
        (i.e. theta=0 & theta=pi) will not be included in the returned
        angular coordinates.
        Inputs: Angular spacings for theta & phi; to 'include_poles' or not
        Outputs: Two (same size) 1D arrays of 'theta' and 'phi'.
    """
    eps = 1e2 * np.finfo('float').eps
    theta, phi = np.meshgrid(np.arange(delta_theta, np.pi - eps, delta_theta),
                             np.arange(0, 2 * np.pi - eps, delta_phi))

    if include_poles:
        return np.r_[0, theta.flatten(), np.pi], np.r_[0, phi.flatten(), 0]
    else:
        return theta.flatten(), phi.flatten()

def equi_dist_angles(N_approx=500):
    """ Create the angular coordinates of a grid with equi-distance spacings
        in 'theta' (polar/colatitudinal) and 'phi' (azimuthal/longitudinal)
        directions.
        Input: N_approx is the target number of points. The algorithm
               aims to return 'N_approx' coordinates but cannot
               guarantee the number to be exact.
        Output: Two (same size) 1-D arrays of theta and phi, the
                colatitudinal and azimuthal angles of the equi-distance
                points on a sphere.
    """
    n_count = 0
    theta_list = []
    phi_list = []
    a = 4 * np.pi / N_approx
    d = np.sqrt(a)
    M_th = np.round(np.pi / d)
    d_theta = np.pi / M_th
    d_phi = a / d_theta
    for m in range(int(M_th)):
        theta = np.pi * (m + 0.5) / M_th
        M_ph = np.round(2 * np.pi * np.sin(theta) / d_phi)
        for n in range(int(M_ph)):
            theta_list.append(theta)
            phi_list.append(2 * np.pi * n / M_ph)
            n_count += 1
    return np.asarray(theta_list), np.asarray(phi_list)

def cta2034a_angles():
    """ Generate the 70 pts angular coordinates for the CTA-2034-A
        spinorama measurement/reporting grid.
        Input: None.
        Output: Two 1-D arrays of length 70.
                'cta_theta' are the polar/colatitudinal angles;
                'cta_phi' are the azimuthal/longitudinal .
                
    """
    cta_theta = (np.pi / 180) * np.r_[0:350:36j, 90:90:34j]
    cta_phi = (np.pi / 180) * np.r_[0:0:36j, 10:170:17j, 190:350:17j]
    return cta_theta, cta_phi

def spherical_grid(n_theta=91, n_phi=91, theta_span=[0, np.pi],
                   phi_span=[np.pi, 2*np.pi]):
    """ Create angular coordinates for a spherical grid  for surface plotting.
        Angular spacings are equally distributed in 'theta'
        (polar/co-latitudinal) and 'phi' (azimuthal/longitudinal) directions.
        Outputs: 2D meshgrid arrays for 'theta' and 'phi'.
    """
    return np.meshgrid(np.linspace(theta_span[0], theta_span[-1], n_theta),
                       np.linspace(phi_span[0], phi_span[-1], n_phi))

def cart_disc_grid(plane='xz', n_r=21, n_ang=91, r_span=[0.5, 1.5],
                   ang_span=[0, 2*np.pi]):
    """ Create a flat hollow disc plotting grid for surface plotting.
        The grid may be on the xy, yz, or xz plane.
        Return three (3) 2D arrays of x, y, z mesh coordinates defining the
        disc surface and is intended for 3D mesh surface plots.
    """
    m_r, m_ang = np.meshgrid(np.linspace(r_span[0], r_span[-1], n_r),
                             np.linspace(ang_span[0], ang_span[-1], n_ang))
    if (plane.lower() == 'xz' or plane.lower() == 'zx'):
        return sph_to_cart(m_r, m_ang, 0)
    elif (plane.lower() == 'yz' or plane.lower() == 'zy'):
        return sph_to_cart(m_r, m_ang, 0.5*np.pi)
    elif (plane.lower() == 'xy' or plane.lower() == 'yx'):
        return sph_to_cart(m_r, 0.5*np.pi, m_ang)
    else:
        raise ValueError("'plane' must be 'xy', 'xz' or 'yz'.")

def sim_sound_cart(x, y, z, omega, pt_srcs, temp=293.15):
    """ Simulate the complex pressure amplitudes at locations 'x', 'y' and
        'z' at frequency 'omega' due to point sources as defined in
        'pt_srcs' defined in Cartesian coordinates in the form:
          [[amplitude1, x1, y1, z1, phase1], [...], ... [...]]
        'temp' is temperature in K.
        Outputs: Array of complex pressure magnitudes, same dimension as 'x'
    """
    c = np.sqrt(1.4 * R_air * temp)
    p = np.zeros(x.shape, dtype=complex)
    for src in pt_srcs:
        ctr_x, ctr_y, ctr_z = src[1], src[2], src[3]
        r = np.sqrt((x - ctr_x)**2 + (y - ctr_y)**2 + (z - ctr_z)**2)
        # Limit r to be at least 1e-5 to avoid singularity
        if r.min() < 1e-5:
            r[r < 1e-5] = 1e-5
        # Sign of the imaginary part needs to be positive to work!
        p += (src[0] / r) * np.exp(1j * (r * omega / c + src[4]))
    return p

def sim_sound_sph(r, theta, phi, omega, pt_srcs, temp=293.15):
    """ Simulate the complex pressure amplitudes at spherical coordinates 'r',
        'theta' and 'phi' at frequency 'omega'. The spherical coordinates are
        usually generated by using the function 'meas_grid'.
        The sound producing point sources are defined in 'pt_srcs' defined
        in Cartesian coordinates in the form:
          [[amplitude1, x1, y1, z1, phase1], [...], ... [...]]
        'temp' is temperature in K.
        Outputs: Array of complex pressure magnitudes, same dimension as
                 'theta'
    """
    x, y, z = sph_to_cart(r, theta, phi)
    return sim_sound_cart(x, y, z, omega, pt_srcs, temp)

def sph_PSI_mix(r, theta, phi, omega, N, temp=293.15):
    """ Compute the mixed field spherical wave expansion function matrix PSI.
        Inputs: The arrays of r, theta, and phi specify the coordinates.
                omega is frequency in rad/s.
                N is the highest order of the coefficients to compute.
                temp is temperature in K.
        Output: Computed spherical wave expansion function matrix PSI in
                complex numbers. The size of the matrix is: M rows by
                2*(N+1)**2 columns, where M = # of coordinates.
    """
    c = np.sqrt(1.4 * R_air * temp)
    if np.size(r) == 1:
        r = np.full_like(theta, r)
    kr = r * omega / c
    PSI_mat = np.empty((theta.size, 2 * (N + 1)**2), dtype='complex')
    for n in range(N + 1):
            for m in range(-n, n + 1):
                j = 2 * (n**2 + n + m)
                PSI_mat[:, j : j + 2] = \
                    np.transpose(np.array([spherical_hn1(n, kr),
                        spherical_jn(n, kr)]) * sph_harm(m, n, phi, theta))
    return PSI_mat

def sph_PSI_int(r, theta, phi, omega, N, temp=293.15):
    """ Compute the interior source only (i.e. anechoic/free field) spherical
        wave expansion function matrix PSI.
        Inputs: The arrays of r, theta, and phi specify the coordinates.
                omega is frequency in rad/s.
                N is the highest order of the coefficients to compute.
                temp is temperature in K.
        Output: Computed spherical wave expansion function matrix PSI in
                complex numbers. The size of the matrix is:
                M rows by (N+1)**2 columns, where M = # of coordinates.
    """
    c = np.sqrt(1.4 * R_air * temp)
    if np.size(r) == 1:
        r = np.full_like(theta, r)
    kr = r * omega / c
    PSI_mat = np.empty((r.size, (N + 1)**2), dtype='complex')
    for n in range(N + 1):
            for m in range(-n, n + 1):
                j = n**2 + n + m
                PSI_mat[:, j] = spherical_hn1(n, kr) \
                                    * sph_harm(m, n, phi, theta)
    return PSI_mat

def sph_PSI_ext(r, theta, phi, omega, N, temp=293.15):
    """ Compute the exterior source only (i.e. external noise only, loudspeaker
        under test is idle) spherical wave expansion function matrix PSI.
        Inputs: The arrays of r, theta, and phi specify the coordinates.
                omega is frequency in rad/s.
                N is the highest order of the coefficients to compute.
                temp is temperature in K.
        Output: Computed spherical wave expansion function matrix PSI in
                complex numbers. The size of the matrix is:
                M rows by (N+1)**2 columns, where M = # of coordinates.
    """
    c = np.sqrt(1.4 * R_air * temp)
    if np.size(r) == 1:
        r = np.full_like(theta, r)
    kr = r * omega / c
    PSI_mat = np.empty((r.size, (N + 1)**2), dtype='complex')
    for n in range(N + 1):
            for m in range(-n, n + 1):
                j = n**2 + n + m
                PSI_mat[:, j] = spherical_jn(n, kr) \
                                    * sph_harm(m, n, phi, theta)
    return PSI_mat

def sph_PSI_csph(theta, phi, N):
    """ Compute the spherical wave expansion function matrix PSI for the case
        when the measurement surface is spherical (i.e. constant r, typ. for
        concentric spherical measurements). In this case, the matrix cells
        only contains the spherical harmonics and is independent of r.
        Mostly used for concentric spherical measurement SFS.
        Inputs: The arrays of theta, and phi specify reconstruction pts.
                N is the highest order of the spherical harmonics to compute.
        Output: Computed spherical wave expansion function matrix PSI in
                complex numbers. The size of the matrix is:
                M rows by (N+1)**2 columns, where M = # of measurement pts.
    """
    PSI_mat = np.empty((theta.shape[0], (N + 1)**2), dtype='complex')
    for n in range(N + 1):
        for m in range(-n, n + 1):
            j = n**2 + n + m
            PSI_mat[:, j] = sph_harm(m, n, phi, theta)
    return PSI_mat

def c_vect_cmtch(r1, r2, omega, temp, N, alpha, beta):
    """ Solve for the weighting coefficient vectors C_mn and D_mn for the
        double concentric spherical measurements case when angular coordinates
        are matched in the two concentric measurements.
        Inputs: r1, r2 are the radii of the two measurement surfaces.
                omega is the angular frequency in rad/s.
                N is the highest order of the spherical harmonics used.
                alpha, beta are the weighting coefficients of the respective
                measurement surfaces.
        Output: C_mn, D_mn are the weighting coefficients. C_mn are for the
                outgoing waves, D_mn are for the standing waves.
    """
    c = np.sqrt(1.4 * R_air * temp)
    k = omega / c
    C_mn = np.empty_like(alpha)
    D_mn = np.empty_like(beta)
    for n in range(N + 1):
        h1n_kr1 = spherical_hn1(n, k * r1)
        h1n_kr2 = spherical_hn1(n, k * r2)
        jn_kr1 = spherical_jn(n, k * r1)
        jn_kr2 = spherical_jn(n, k * r2)
        for m in range(-n, n + 1):
            l = n**2 + n + m
            C_mn[l], D_mn[l] = np.linalg.solve(
                        np.array([[h1n_kr1, jn_kr1], [h1n_kr2, jn_kr2]]),
                        np.array([alpha[l], beta[l]]))
    return C_mn, D_mn

def point_plot_sph(grid, press, v_range=np.array([0, 1]),
                   sound_sources=None):
    """ 3-D point plot of sound pressures. Each data point location (in
        spherical coordinates) is indicated by a sphere. Size of sphere
        and color is scaled to the magnitude of the complex sound pressure
        amplitude.
        Inputs: 'grid' are the location coordinates in spherical coordinates.
                'press' are the complex pressure amplitudes,
                Min and max range for the color scale is v_min='v_range[0]'
                  and v_max='v_range[-1]'.
                If 'sound_source' is given, the locations are plotted as 3-D
                  crosses.
    """
    grid_r, grid_theta, grid_phi = grid[0], grid[1], grid[2]
    if grid_r.shape == press.shape:
        pts_x, pts_y, pts_z = sph_to_cart(grid_r, grid_theta, grid_phi)
    elif grid_r.shape == (2, ):
        pts_x, pts_y, pts_z = \
            sph_to_cart(np.r_[np.full_like(grid_theta, grid_r[0]),
                              np.full_like(grid_theta, grid_r[1])],
                        np.r_[grid_theta, grid_theta],
                        np.r_[grid_phi, grid_phi])
    else:
        raise ValueError("Grid shape: {}. ".format(grid_r.shape) +
                         "Must be (2, ) or the same as grid_theta.")
    
    pts_press = np.abs(press.flatten().reshape(pts_x.shape))
    scale_factor = np.minimum(np.maximum(0.05/pts_press.max()**2, 1e-5),
                              v_range[-1])
    mlab.points3d(pts_x, pts_y, pts_z, pts_press, line_width=0.2,
                  scale_factor=scale_factor, colormap='viridis',
                  scale_mode='scalar', resolution=16, mode='sphere',
                  vmax=v_range[-1], vmin=v_range[0])
    axes_ext = 1.1 * np.array([-grid_r.max(), grid_r.max()])
    mlab.plot3d(axes_ext, np.array([0, 0]), np.array([0, 0]), tube_radius=1e-3,
                color=(1, 0, 0), representation='wireframe')
    mlab.plot3d(np.array([0, 0]), axes_ext, np.array([0, 0]), tube_radius=1e-3,
                color=(0, 1, 0), representation='wireframe')
    mlab.plot3d(np.array([0, 0]), np.array([0, 0]), axes_ext, tube_radius=1e-3,
                color=(0, 0, 1), representation='wireframe')
    if not(sound_sources is None):
        for src in sound_sources:
            mlab.points3d(src[1], src[2], src[3], color=(1.0, 0.0, 1.0),
                          line_width=3.0, scale_factor=0.04, resolution=8,
                          mode='axes')

def surf_plot_sph(grid, press, v_range=np.array([0, 1]), sound_sources=None):
    """ 3-D surface plot of sound pressures. Color is scaled to the magnitude
        of the complex sound pressure amplitude.
        Inputs: 'grid' are the location coordinates in spherical coordinates.
                'press' are the complex pressure amplitudes,
                Min and max range for the color scale is v_min='v_range[0]'
                  and v_max='v_range[-1]'.
                If 'sound_source' is given, the locations are plotted as 3-D
                  crosses.
    """
    grid_r, grid_theta, grid_phi = grid[0], grid[1], grid[2]
    if grid_r.shape == press.shape:
        pts_x, pts_y, pts_z = sph_to_cart(grid_r, grid_theta, grid_phi)
    elif grid_r.shape == (2, ):
        pts_x, pts_y, pts_z = sph_to_cart(
            np.r_[np.full_like(grid_theta, grid_r[0]),
                  np.full_like(grid_theta, grid_r[1])],
            np.r_[grid_theta, grid_theta],
            np.r_[grid_phi, grid_phi])
    else:
        raise ValueError("Grid shape: {}. ".format(grid_r.shape) +
                         "Must be (2, ) or the same as grid_theta.")
    
    pts_press = np.abs(press.flatten().reshape(pts_x.shape))
    mlab.mesh(pts_x, pts_y, pts_z, scalars=pts_press, colormap='viridis',
              vmax=v_range[-1], vmin=v_range[0])
    axes_ext = 1.1 * np.array([-grid_r.max(), grid_r.max()])
    mlab.plot3d(axes_ext, np.array([0, 0]), np.array([0, 0]), tube_radius=1e-3,
                color=(1, 0, 0), representation='wireframe')
    mlab.plot3d(np.array([0, 0]), axes_ext, np.array([0, 0]), tube_radius=1e-3,
                color=(0, 1, 0), representation='wireframe')
    mlab.plot3d(np.array([0, 0]), np.array([0, 0]), axes_ext, tube_radius=1e-3,
                color=(0, 0, 1), representation='wireframe')
    if not(sound_sources is None):
        for src in sound_sources:
            mlab.points3d(src[1], src[2], src[3], color=(1.0, 0.0, 1.0),
                          line_width=3.0, scale_factor=0.04, resolution=8,
                          mode='axes')

def press_plot_points(speaker, which='recon', v_range=np.array([0, 1]),
                      sound_sources=None, save_fig=False, res=None):
    """ 3-D plot of sound pressures with plot point location indicated
        by a sphere. Size of sphere and color are scaled to the magnitude
        of the complex sound pressure amplitude.
        Inputs: 'speaker' is the speaker object.
                'which' - if 'meas', plot the measurements
                          if 'recon' plot the reconstruction pressures
                Min and max range for the color scale is v_min='v_range[0]'
                  and v_max='v_range[-1]'.
                If 'sound_source' is given, the locations are plotted as 3-D
                  crosses.
                If 'save_fig' is True, figure is saved either as
                  'p_meas.png' or 'p_recon.png'.
                If residual 'res' is provided, it is display on the plot.
                  Applicable only when 'which is 'recon'.
    """
    if which.lower() == 'meas':
        grid = speaker.meas.grid
        press = speaker.meas.press
        n_points = press.size
    elif which.lower() == 'recon':
        grid = speaker.recon.grid
        press = speaker.recon.press
        N = int(np.sqrt(speaker.coeff_vect[0].size)) - 1
    else:
        raise ValueError("Invalid 'which', must be 'meas' or 'recon'.")
    
    fig = mlab.figure(1, bgcolor=(1, 1, 1), fgcolor=(0, 0, 0), size=(800, 600))
    mlab.clf()
    point_plot_sph(grid, press, v_range, sound_sources)
    
    if which.lower() == 'meas':
        mlab.text(0.15, 0.92, '# Meas={:d}'.format(n_points),
                  width=0.20)
        mlab.text(0.15, 0.88, 'Freq={:.0f} Hz'.format(speaker.meas.freq),
                  width=0.14)
        fname = 'p_meas_F{:.0f}.png'.format(speaker.meas.freq)
    else:
        mlab.text(0.15, 0.92, 'Order={:2d}'.format(N), width=0.12)
        mlab.text(0.15, 0.88, 'Freq={:.0f} Hz'.format(speaker.meas.freq),
                  width=0.14)
        if not(res is None):
            mlab.text(0.15, 0.84, 'Res={:.3e}'.format(res), width=0.15)
        fname = 'p_recon_F{:.0f}_N{:02d}.png'.format(speaker.meas.freq, N)

    axes_ext = 1.1 * np.array([-grid[0].max(), grid[0].max()])
    mlab.view(azimuth=45, elevation=70, distance=3.85*axes_ext[-1],
              focalpoint=(0.15, -0.10, 0.05))
    mlab.colorbar(orientation='vertical', nb_labels=6)
    mlab.orientation_axes()
    if save_fig:
        mlab.savefig(fname)
    mlab.show()

class Speaker(object):
    def __init__(self):
        """ Loudspeaker under test class.
        """
        self.debug_mode = False
        class measurements():
            pass
        self.meas = measurements()
        self.meas.method = None
        self.meas.temp = None
        self.meas.freq = None
        self.meas.omega = None
        self.meas.grid_r = None
        self.meas.grid_theta = None
        self.meas.grid_phi = None
        self.meas.grid = (self.meas.grid_r, self.meas.grid_theta,
                          self.meas.grid_phi)
        self.meas.press = None
        self.meas.PSI_mat = np.array([[], []])
        
        class reconstruction():
            pass
        self.recon = reconstruction()
        self.recon.grid_r = None
        self.recon.grid_theta = None
        self.recon.grid_phi = None
        self.recon.grid = (self.recon.grid_r, self.recon.grid_theta,
                           self.recon.grid_phi)
        self.recon.PSI_mat = np.array([[], []])
        self.recon.press = None

        self.regress_params = None
        self.coeff_vect = None

    def set_meas_grid(self, r, theta, phi, method):
        """ Set the measurement method and grid coordinates (spherical)
        """
        self.meas.grid_r = np.asarray(r)
        self.meas.grid_theta = np.asarray(theta)
        self.meas.grid_phi = np.asarray(phi)
        self.meas.grid = (self.meas.grid_r, self.meas.grid_theta,
                          self.meas.grid_phi)
        if method.lower() == 'asph':
            self.meas.method = 'asph'
        elif method.lower() == 'cmtch':
            self.meas.method = 'cmtch'
        elif method.lower() == 'cdiss':
            self.meas.method = 'cdiss'
        else:
            raise ValueError("Invalid measurement method, must be "
                             + "'asph', 'cmcth' or 'cdiss'.")

    def set_meas_press(self, freq, press, temp=293.15):
        """ Set the measurements - complex sound pressure amplitudes
            Inputs: 'freq' is frequency
                    'press' is an array of complex pressure amplitudes
                    'temp' is ambient air temperature during measurements
        """
        self.meas.freq = freq
        self.meas.omega = 2 * np.pi * freq
        self.meas.press = np.asarray(press)
        self.meas.temp = temp

    def set_recon_grid(self, r, theta, phi):
        """ Set the reconstruction grid coordinates (spherical)
        """
        self.recon.grid_r = np.asarray(r)
        self.recon.grid_theta = np.asarray(theta)
        self.recon.grid_phi = np.asarray(phi)
        self.recon.grid = np.asarray((self.recon.grid_r, self.recon.grid_theta,
                                      self.recon.grid_phi))
        
    def set_recon_press(self, press):
        """ Set the reconstruction complex sound amplitudes
        """
        self.recon.press = np.asarray(press)

    def gen_PSI_meas_asph(self, N, PSI_mat=None):
        """ Compute the PSI matrix for aspherical measurements and place it
            into 'self.meas.PSI_mat'.
            If 'PSI_mat' is not give, existing 'self.meas.PSI_mat' is used.
            If 'PSI_mat' has more columns than if a new PSI matrix is to be
            computed with N, then the input 'PSI_mat' is placed into
            'self.meas.PSI_mat' as it is assumed that the input 'PSI_mat'
            was computed previously with a higher 'N', and therefore the
            required elements (columns) are already computed and present.
            Inputs: 'N' is the highest order spherical harmonics to compute to.
                    'PSI_mat' is an existing PSI matrix for the new PSI matrix
                    to build from, or None to use existing 'self.meas.PSI_mat'.
            Output: None.
                    If PSI_mat is given and is of sufficient dimension for
                    order N, it replaces 'self.meas.PSI_mat'.
                    If PSI_mat is not given, and 'self.meas.PSI_mat' is of
                    sufficient dimension for order N, it is left untouched.
                    Otherwise, a new PSI_mat is computed and it replaces
                    'self.meas.PSI_mat'.
        """
        if PSI_mat is None:
            # Use the existing measurement PSI matrix in the Speaker object.
            PSI_mat = self.meas.PSI_mat
        else:
            # Use the PSI matrix given
            PSI_mat = np.asarray(PSI_mat)
            
        if self.debug_mode:
            print('PSI_mat shape {}'.format(np.shape(PSI_mat)))
        
        if ((PSI_mat.shape[1] < 1) or
            (PSI_mat.shape[0] != self.meas.grid_theta.shape[0])):
            # If empty PSI_mat or PSI_mat dim not match # of press
            # measurements, need to generate PSI_mat from scratch
            N_start = 0
            if self.debug_mode:
                print('Cond 1a: N_start = {}'.format(N_start))
            Mat = np.empty((self.meas.grid_theta.shape[0], 2*(N + 1)**2),
                            dtype='complex')
        elif (PSI_mat.shape[1] < 2*(N + 1)**2):
            # More columns need to be added to PSI_mat
            N_start = int(np.sqrt(PSI_mat.shape[1] / 2))
            if self.debug_mode:
                print('Cond 2a: N_start = {}'.format(N_start))
            Mat = np.empty((self.meas.grid_theta.shape[0],
                           2*((N + 1)**2 - N_start**2)), dtype='complex')
        else:
            # PSI_mat is of sufficient dimension, assume it is OK!
            if self.debug_mode:
                print('Cond 3a: PSI mat is assumed OK!')
            self.meas.PSI_mat = PSI_mat
            return None

        j = 0
        kr = (self.meas.grid_r * self.meas.omega
              / np.sqrt(1.4 * R_air * self.meas.temp))
        for n in range(N_start, N + 1):
            for m in range(-n, n + 1):
                sph_harm_mn = sph_harm(m, n, self.meas.grid_phi,
                                       self.meas.grid_theta)
                Mat[:, j : j + 2] = \
                    np.transpose(np.array([spherical_hn1(n, kr) * sph_harm_mn,
                                           spherical_jn(n, kr) * sph_harm_mn]))
                j += 2
            
        if N_start == 0:
            self.meas.PSI_mat = Mat
        else:
            self.meas.PSI_mat = np.column_stack((PSI_mat, Mat))
        return None

    def gen_PSI_meas_cmtch(self, N, PSI_mat=None):
        """ Compute the PSI matrix for concentric spherical measurements 
            with matching angular coordinates, and place it into
            'self.PSI_mat_meas'.
            If input PSI_matrix is not give, 'self.meas.PSI_mat' is used.
            If input PSI matrix has more columns than if a new one is to be
            computed with N, then the input PSI matrix is placed into
            'self.meas.PSI_mat' as it is assumed that the input 'PSI_mat'
            was computed previously with a higher 'N', and therefore the
            required elements (columns) already computed and present.
            Inputs: N is the highest order spherical harmonics to compute to.
                    'PSI_mat' is an existing PSI matrix for the new PSI matrix
                    to build from, or None to use existing 'self.meas.PSI_mat'.
            Output: None.
                    If PSI_mat is given and is of sufficient dimension for
                    order N, it replaces 'self.meas.PSI_mat'.
                    If PSI_mat is not given, and 'self.meas.PSI_mat' is of
                    sufficient dimension for order N, it is left untouched.
                    Otherwise, a new PSI_mat is computed and it replaces
                    'self.meas.PSI_mat'.
        """
        if PSI_mat is None:
            # Use the existing measurement PSI matrix in the Speaker object.
            PSI_mat = self.meas.PSI_mat
        else:
            # Use the PSI matrix given
            PSI_mat = np.asarray(PSI_mat)
            
        if self.debug_mode:
            print('PSI_mat shape {}'.format(np.shape(PSI_mat)))
        
        if ((PSI_mat.shape[1] < 1) or
            (PSI_mat.shape[0] != self.meas.grid_theta.shape[0])):
            # If empty PSI_mat or PSI_mat dim not match # of press
            # measurements, need to generate PSI_mat from scratch
            N_start = 0
            if self.debug_mode:
                print('Cond 1c: N_start = {}'.format(N_start))
            Mat = np.empty((self.meas.grid_theta.shape[0], (N + 1)**2),
                               dtype='complex')
        elif (PSI_mat.shape[1] < (N + 1)**2):
            # More columns need to be added to PSI_mat
            N_start = int(np.sqrt(PSI_mat.shape[1]))
            if self.debug_mode:
                print('Cond 2c: N_start = {}'.format(N_start))
            Mat = np.empty((self.meas.grid_theta.shape[0],
                           (N + 1)**2 - N_start**2), dtype='complex')
        else:
            # PSI_mat is of sufficient dimension, assume it is OK!
            if self.debug_mode:
                print('Cond 3c: PSI mat is AOK')
            self.meas.PSI_mat = PSI_mat
            return None

        j = 0
        for n in range(N_start, N + 1):
            for m in range(-n, n + 1):
                Mat[:, j] = sph_harm(m, n, self.meas.grid_phi,
                                     self.meas.grid_theta)
                j += 1
            
        if N_start == 0:
            self.meas.PSI_mat = Mat
        else:
            self.meas.PSI_mat = np.column_stack((PSI_mat, Mat))
        return None

    def coeff_solve(self, N, solver='lstsq'):
        """ Solve for the spherical wave expansion function coefficients.
            Methods are:
            'lstsq' - standard least squares (numpy.linalg.lstsq).
            'tik' - modified Tikhonov regularization with generalized
                    cross validation
            Results are put in 'self.coeff_vect' as [[C_mn], [D_mn]]
        """
        if self.meas.method == 'asph':
            # Aspherical measurements
            # Build PSI matrix
            self.gen_PSI_meas_asph(N=N)
            # Solve for coefficients
            J = 2 * (N + 1)**2
            if solver.lower() =='lstsq':
                soln = np.linalg.lstsq(self.meas.PSI_mat[:, 0:J],
                                       self.meas.press, rcond=None)
            elif solver.lower() =='tik':
                raise ValueError("Method 'tik' not implemented yet!")
            else:
                raise ValueError("Invalid method, must be 'lstsq' or 'tik'.")
            C_mn = soln[0][::2]
            D_mn = soln[0][1::2]
            if soln[1].size == 0:
                res = np.nan
            else:
                res = soln[1][0]
            rank = soln[2]
            cond_num = soln[3][0] / soln[3][-1]
        elif self.meas.method == 'cmtch':
            # Concentric spherical matched angular coordinates measurements
            # Build PSI matrix
            self.gen_PSI_meas_cmtch(N=N)
            # Solve for coeff_vect
            J = (N + 1)**2
            if solver.lower() =='lstsq':
                soln1 = np.linalg.lstsq(self.meas.PSI_mat[:, 0:J],
                                    self.meas.press[0], rcond=None)
                soln2 = np.linalg.lstsq(self.meas.PSI_mat[:, 0:J],
                                    self.meas.press[1], rcond=None)
                C_mn, D_mn = c_vect_cmtch(self.meas.grid_r[0],
                                          self.meas.grid_r[1],
                                          self.meas.omega, self.meas.temp,
                                          N, soln1[0], soln2[0])
                if (soln1[1].size == 0) or (soln2[1].size == 0):
                    res = np.nan
                else:
                    res = soln1[1][0] + soln2[1][0]
                rank = np.minimum(soln1[2], soln2[2])
                cond_num = np.maximum(soln1[3][0] / soln1[3][-1],
                                      soln2[3][0] / soln2[3][-1])
            elif solver.lower() =='tik':
                raise ValueError("Method 'tik' not implemented yet!")
            else:
                raise ValueError("Invalid method, must be 'lstsq' or 'tik'.")
        elif self.meas.method == 'cdiss':
            # Concentric spherical measurements dissimilar angular coordinates
            raise ValueError("Sorry, 'cdiss' method not implemented!")
        else:
            raise ValueError("Invalid measurement method!")

        self.coeff_vect = np.array([C_mn, D_mn])
        self.regress_params = (res, rank, cond_num)
        return None

    def recon_press(self, recon_type='int', recon_grid=None):
        """ Reconstruct sound pressure. The coefficient vector must already
            be available.
        """
        if recon_grid is None:
            recon_grid = self.recon.grid
            shape = self.recon.grid_theta.shape
        elif ((recon_grid[0] is None) or (recon_grid[1] is None)
            or (recon_grid[2] is None)):
            raise ValueError("Reconstruction grid cannot be empty!")
        else:
            shape = recon_grid[1].shape
            self.set_recon_grid(*recon_grid)

        recon_grid = (self.recon.grid[0].flatten(),
                      self.recon.grid[1].flatten(),
                      self.recon.grid[2].flatten())
        # print('Recon grid shape:{}'.format(np.shape(recon_grid)), end=' ')
        # print('Recon type:{}'.format(recon_type))
        N = int(np.sqrt(self.coeff_vect[0].size)) - 1
        if recon_type == 'int':
            self.recon.PSI_mat = sph_PSI_int(*recon_grid, self.meas.omega, N,
                                             self.meas.temp)
            self.recon.press = np.dot(self.recon.PSI_mat, self.coeff_vect[0])
        elif recon_type == 'mix':
            self.recon.PSI_mat = sph_PSI_mix(*recon_grid, self.meas.omega, N,
                                             self.meas.temp)
            self.recon.press = np.dot(self.recon.PSI_mat,
                np.column_stack((self.coeff_vect[0],
                                 self.coeff_vect[1])).flatten())
        elif recon_type == 'ext':
            self.recon.PSI_mat = sph_PSI_ext(*recon_grid, self.meas.omega, N,
                                             self.meas.temp)
            self.recon.press = np.dot(self.recon.PSI_mat, self.coeff_vect[1])
        
        self.recon.press.resize(shape)